"""Top-level package for CARSpy."""
from .line_strength import LineStrength
from .cars_synth import CarsSpectrum
from .cars_fit import CarsFit
from .utils import *

__author__ = """Zhiyao Yin"""
__email__ = 'zhiyao.yin@dlr.de'
__version__ = '0.6.0'
